// MultithreadingHazards.cs
//
// This very simply application just demonstrates (1) the hazards associated with accessing
// controls in a Windows Forms app from the incorrect thread and (2) the work around.
//
// Here's a description of the UI for this application:
//
//  "Enter increment count:"            The value entered here indicates how many
//                                      requests to make of the thread pool to call
//                                      the AddToactualValue method, which
//                                      updates the 'actualValue' label shown
//                                      in the "Results" groupbox.
//
//  "Increment by how much?"            The value entered here indicates the value
//                                      that will be added to actualValue by
//                                      AddToactualValue.
//
//  "Test and react to InvokeRequired?" Controls whether or not InvokeRequired
//                                      and BeginInvoke are used appropriately to
//                                      prevent multithreaded errors.
//
//  "Expected results"                  The value that should result from pressing
//                                      the "Go!" button if no errors were encountered.
//
//  "Actual results"                    The actual value that resulted from the way
//                                      AddToactualValue was implemented.
//
// For example, if you enter 100 for the increment count and 1 for the 'by how much' value,
// then the thread pool will call AddToactualValue 100 times, and each such call
// will add 1 to actualValue.  Assuming "Actual results" was 0 before you pressed
// the "Go!" button, it should be equal to "100" after the calls to AddToactualValue
// complete.
//
// If "Test and react to InvokeRequired?" is not checked, then the outcome of the
// operation cannot be guaranteed, and you'll see that the actual results don't match
// the expected results always (sometimes they do - indicating you were lucky; sometimes
// they don't - which is what you have to assume will happen when you can least afford it to).
//
// If "Test and react to InvokeRequired?" is checked, then InvokeRequired & BeginInvoke
// will be used appropriately to ensure that the form's controls are always accessed from
// the correct thread.  Specifically, it means that the read/modify/write sequence will
// be performed atomically in the context of only the form thread, preventing the race
// condition that results from doing these operations on different threads (resulting
// in "lost" updates in this example).
//
// Refer to goButton_Click and AddToActualValue for the relevant code.
//
// Mike Woodring
// http://staff.develop.com/woodring
//
using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.Threading;

namespace MultithreadingHazards
{
	/// <summary>
	/// Summary description for Form1.
	/// </summary>
	public class Form1 : System.Windows.Forms.Form
	{
        bool testThread = false;

        private System.Windows.Forms.Label label1;
        private System.Windows.Forms.Button goButton;
        private System.Windows.Forms.Label label2;
        private System.Windows.Forms.TextBox incrementCount;
        private System.Windows.Forms.Label actualValue;
        private System.Windows.Forms.Button resetButton;
        private System.Windows.Forms.GroupBox groupBox1;
        private System.Windows.Forms.TextBox incrementValue;
        private System.Windows.Forms.Label label3;
        private System.Windows.Forms.Label label4;
        private System.Windows.Forms.Label expectedResults;
        private System.Windows.Forms.CheckBox reactToInvokeRequired;
		/// <summary>
		/// Required designer variable.
		/// </summary>
		private System.ComponentModel.Container components = null;

		public Form1()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
            this.label1 = new System.Windows.Forms.Label();
            this.incrementCount = new System.Windows.Forms.TextBox();
            this.goButton = new System.Windows.Forms.Button();
            this.label2 = new System.Windows.Forms.Label();
            this.actualValue = new System.Windows.Forms.Label();
            this.resetButton = new System.Windows.Forms.Button();
            this.groupBox1 = new System.Windows.Forms.GroupBox();
            this.expectedResults = new System.Windows.Forms.Label();
            this.label4 = new System.Windows.Forms.Label();
            this.incrementValue = new System.Windows.Forms.TextBox();
            this.label3 = new System.Windows.Forms.Label();
            this.reactToInvokeRequired = new System.Windows.Forms.CheckBox();
            this.groupBox1.SuspendLayout();
            this.SuspendLayout();
            // 
            // label1
            // 
            this.label1.Location = new System.Drawing.Point(8, 16);
            this.label1.Name = "label1";
            this.label1.Size = new System.Drawing.Size(120, 23);
            this.label1.TabIndex = 0;
            this.label1.Text = "Enter increment count:";
            // 
            // incrementCount
            // 
            this.incrementCount.Location = new System.Drawing.Point(152, 16);
            this.incrementCount.Name = "incrementCount";
            this.incrementCount.Size = new System.Drawing.Size(48, 20);
            this.incrementCount.TabIndex = 1;
            this.incrementCount.Text = "100";
            // 
            // goButton
            // 
            this.goButton.Location = new System.Drawing.Point(208, 16);
            this.goButton.Name = "goButton";
            this.goButton.TabIndex = 2;
            this.goButton.Text = "Go!";
            this.goButton.Click += new System.EventHandler(this.goButton_Click);
            // 
            // label2
            // 
            this.label2.Location = new System.Drawing.Point(32, 48);
            this.label2.Name = "label2";
            this.label2.Size = new System.Drawing.Size(80, 23);
            this.label2.TabIndex = 3;
            this.label2.Text = "Actual results:";
            // 
            // actualValue
            // 
            this.actualValue.Location = new System.Drawing.Point(136, 48);
            this.actualValue.Name = "actualValue";
            this.actualValue.Size = new System.Drawing.Size(48, 23);
            this.actualValue.TabIndex = 3;
            this.actualValue.Text = "0";
            // 
            // resetButton
            // 
            this.resetButton.Location = new System.Drawing.Point(192, 48);
            this.resetButton.Name = "resetButton";
            this.resetButton.TabIndex = 4;
            this.resetButton.Text = "Reset";
            this.resetButton.Click += new System.EventHandler(this.resetButton_Click);
            // 
            // groupBox1
            // 
            this.groupBox1.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                                    this.expectedResults,
                                                                                    this.label4,
                                                                                    this.resetButton,
                                                                                    this.label2,
                                                                                    this.actualValue});
            this.groupBox1.Location = new System.Drawing.Point(8, 104);
            this.groupBox1.Name = "groupBox1";
            this.groupBox1.Size = new System.Drawing.Size(280, 80);
            this.groupBox1.TabIndex = 5;
            this.groupBox1.TabStop = false;
            this.groupBox1.Text = "Test Results";
            // 
            // expectedResults
            // 
            this.expectedResults.Location = new System.Drawing.Point(136, 24);
            this.expectedResults.Name = "expectedResults";
            this.expectedResults.Size = new System.Drawing.Size(48, 23);
            this.expectedResults.TabIndex = 6;
            this.expectedResults.Text = "0";
            // 
            // label4
            // 
            this.label4.Location = new System.Drawing.Point(32, 24);
            this.label4.Name = "label4";
            this.label4.Size = new System.Drawing.Size(104, 23);
            this.label4.TabIndex = 5;
            this.label4.Text = "Expected results:";
            // 
            // incrementValue
            // 
            this.incrementValue.Location = new System.Drawing.Point(152, 48);
            this.incrementValue.Name = "incrementValue";
            this.incrementValue.Size = new System.Drawing.Size(48, 20);
            this.incrementValue.TabIndex = 7;
            this.incrementValue.Text = "1";
            // 
            // label3
            // 
            this.label3.Location = new System.Drawing.Point(8, 48);
            this.label3.Name = "label3";
            this.label3.Size = new System.Drawing.Size(136, 23);
            this.label3.TabIndex = 6;
            this.label3.Text = "Increment by how much?";
            // 
            // reactToInvokeRequired
            // 
            this.reactToInvokeRequired.Location = new System.Drawing.Point(48, 72);
            this.reactToInvokeRequired.Name = "reactToInvokeRequired";
            this.reactToInvokeRequired.Size = new System.Drawing.Size(200, 24);
            this.reactToInvokeRequired.TabIndex = 9;
            this.reactToInvokeRequired.Text = "Test and react to InvokeRequired?";
            // 
            // Form1
            // 
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.ClientSize = new System.Drawing.Size(296, 187);
            this.Controls.AddRange(new System.Windows.Forms.Control[] {
                                                                          this.reactToInvokeRequired,
                                                                          this.incrementValue,
                                                                          this.label3,
                                                                          this.groupBox1,
                                                                          this.goButton,
                                                                          this.incrementCount,
                                                                          this.label1});
            this.Name = "Form1";
            this.Text = "Multithreading Hazards";
            this.groupBox1.ResumeLayout(false);
            this.ResumeLayout(false);

        }
		#endregion

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main() 
		{
			Application.Run(new Form1());
		}

        private void goButton_Click(object sender, System.EventArgs e)
        {
            // Make a note of whether we should use InvokeRequired & BeginInvoke
            // inside AddToActualValue;
            // 
            testThread = reactToInvokeRequired.Checked;

            int numIncrements = int.Parse(incrementCount.Text);
            int incValue = int.Parse(incrementValue.Text);

            // Update the predicted/expected value of the "Actual results"
            // label assuming no errors in AddtoactualValue were encountered.
            //
            int expectedValue = int.Parse(actualValue.Text) + (numIncrements * incValue);
            expectedResults.Text = expectedValue.ToString();

            // Fire off numIncrements requests to the thread pool to
            // invoke AddToactualValue.  The value to increment
            // the control by is passed as the 'state' parameter to
            // QueueUserWorkItem.
            //
            WaitCallback incrementProc = new WaitCallback(AddToActualValue);

            for( int n = 0; n < numIncrements; n++ )
            {
                ThreadPool.QueueUserWorkItem(incrementProc, incValue);
            }
        }

        void AddToActualValue( object state )
        {
            if( testThread )
            {
                // Determine if this code has been called on the "wrong" thread,
                // and, if so, marshal a request to the "right" thread (along with a copy of
                // the parameters that were passed to this method), and then bail out
                // without doing anything else.  This method will subsequently be called
                // again from the 'right' thread.
                //
                if( InvokeRequired )
                {
                    BeginInvoke( new WaitCallback(AddToActualValue),
                                new object[]{ state } );
                    return;
                }
            }

            // Read current value of actualValue label.
            //
            int curValue = int.Parse(actualValue.Text);

            // Add 1 to current value.
            //
            curValue += (int)state;

            // Store updated value back into actualValue label.
            //
            actualValue.Text = curValue.ToString();
        }

        private void resetButton_Click(object sender, System.EventArgs e)
        {
            actualValue.Text = expectedResults.Text = "0";
        }
	}
}
